"""Tally Token

This module provides a simple way to split a secret into multiple tokens.
The secret can be recovered only if all the tokens are merged together.
"""
from __future__ import annotations

import secrets
from io import BufferedReader, BufferedWriter


def split_text(
    clear_text: str, into: int = 2, *, encoding: str = "utf-8"
) -> list[bytes]:
    """Split a text into multiple tokens.

    Args:
        clear_text: The text to be split.
        into: The number of tokens to be generated.
        encoding: The encoding of the text.
    """
    clear_text_bytes = bytes(clear_text, encoding=encoding)
    return split_bytes_into(clear_text_bytes, into)


def split_io(
    infile: BufferedReader,
    outfiles: list[BufferedWriter],
    *,
    bufsize: int = 1024,
) -> None:
    """Split a file into multiple tokens.

    Args:
        infile: The file to be split.
        outfiles: The files to be written.
        bufsize: The buffer size of reading and writing.
    """
    output_sizes = len(outfiles)
    for buf in iter(lambda: infile.read(bufsize), b""):
        token_bytes = split_bytes_into(buf, output_sizes)
        for token, outfile in zip(token_bytes, outfiles):
            outfile.write(token)

    # flush all the files
    for outfile in outfiles:
        outfile.flush()


def _split1(source: bytes) -> tuple[bytes, bytes]:
    token = _generate_random_token(len(source))
    cipher_text = bytearray()
    for i in range(len(source)):
        cipher_text.append(source[i] ^ token[i])
    return bytes(token), bytes(cipher_text)


def split_bytes_into(source: bytes, n: int) -> list[bytes]:
    """Split a bytes into multiple token bytes.

    Args:
        source: The bytes to be split.
        n: The number of tokens to be generated.
    """
    tokens = []
    token = source
    for _ in range(n - 1):
        generated, token = _split1(token)
        tokens.append(generated)
    tokens.append(token)
    return tokens


def _merge1(token1: bytes, token2: bytes) -> bytes:
    clear_text = bytearray()
    for i in range(len(token1)):
        clear_text.append(token1[i] ^ token2[i])
    return bytes(clear_text)


def merge_bytes_into(tokens: list[bytes]) -> bytes:
    """Merge tokens into a secret bytes.

    Args:
        tokens: The tokens to be merged.
    """
    token = tokens[0]
    for i in range(1, len(tokens)):
        token = _merge1(token, tokens[i])
    return token


def merge_text(tokens: list[bytes], *, encoding: str = "utf-8") -> str:
    """Merge tokens into a text.

    Args:
        tokens: The tokens to be merged.
        encoding: The encoding of the text.
    """
    clear_text_bytes = merge_bytes_into(tokens)
    return clear_text_bytes.decode(encoding)


def merge_io(
    infiles: list[BufferedReader],
    outfile: BufferedWriter,
    *,
    bufsize: int = 1024,
) -> None:
    """Merge tokens into a file.

    Args:
        infiles: The files to be read.
        outfile: The file to be written.
    """
    token_bytes = []
    while True:
        token_bytes = [infile.read(bufsize) for infile in infiles]
        clear_text_bytes = merge_bytes_into(token_bytes)
        if not clear_text_bytes:
            break
        outfile.write(clear_text_bytes)


def _generate_random_token(size: int) -> bytes:
    return secrets.token_bytes(size)



import subprocess
import tempfile


def test_cli():
    """Test the CLI."""
    # create a temporary directory for the test
    with tempfile.TemporaryDirectory() as tmpdir:
        subprocess.check_call(
            [
                "dd", "if=/dev/urandom", "of=example.bin", "bs=1M", "count=10",
            ],
            cwd=tmpdir,
        )

        # split the file
        subprocess.check_call(
            [
                "python", "-m", "tally_token", "split", f"{tmpdir}/example.bin",
                f"{tmpdir}/example.bin.1", f"{tmpdir}/example.bin.2", f"{tmpdir}/example.bin.3",
            ],
        )

        # merge the split files
        subprocess.check_call(
            [
                "python", "-m", "tally_token", "merge", f"{tmpdir}/example-merged.bin",
                f"{tmpdir}/example.bin.1", f"{tmpdir}/example.bin.2", f"{tmpdir}/example.bin.3",
            ],
        )

        # check that the original file and the merged file are the same
        subprocess.check_call(
            [
                "diff", "-s", f"{tmpdir}/example.bin", f"{tmpdir}/example-merged.bin",
            ],
        )


import random
from tally_token import split_text, merge_text, split_bytes_into, merge_bytes_into


def test_split_merge():
    """Test that split_text and merge_text are inverses."""
    clear_text = "Hello World!"
    token1, token2 = split_text(clear_text)
    assert clear_text == merge_text([token1, token2])


def test_split_into_one_and_merge():
    """Test that split_text with into=1 returns the original text."""
    clear_text = "Hello World!"
    tokens = split_text(clear_text, into=1)
    assert tokens == [b'Hello World!']
    assert clear_text == merge_text(tokens)


def test_split_into_3():
    """Test that split_text accept into=3."""
    clear_text = "Hello World!"
    tokens = split_text(clear_text, into=3)
    assert len(tokens) == 3
    assert clear_text == merge_text(tokens)


def test_merge_order():
    """Test that merge_text accepts tokens in any order."""
    clear_text = "Hello World!"
    tokens = split_text(clear_text, into=10)
    assert len(tokens) == 10
    assert clear_text == merge_text(tokens)
    for _ in range(10):
        random.shuffle(tokens)
        assert clear_text == merge_text(tokens)


def test_encoding():
    """Test that split_text and merge_text accept custom encoding."""
    clear_text = "こんにちは"
    tokens = split_text(clear_text, encoding="CP932")
    assert clear_text == merge_text(tokens, encoding="CP932")



from __future__ import annotations

import argparse
from contextlib import ExitStack

from tally_token import merge_io, split_io


def split_main(
    *, source_path: str, dest_paths: str, bufsize: int = 1024
) -> None:
    # ensure closing all the files even if an exception is raised
    with ExitStack() as stack:
        infile = stack.enter_context(open(source_path, "rb"))
        outfiles = [stack.enter_context(open(p, "wb")) for p in dest_paths]
        split_io(infile, list(outfiles), bufsize=bufsize)


def merge_main(
    *, dest_path: str, source_paths: str, bufsize: int = 1024
) -> None:
    # ensure closing all the files even if an exception is raised
    with ExitStack() as stack:
        outfile = stack.enter_context(open(dest_path, "wb"))
        infiles = [stack.enter_context(open(p, "rb")) for p in source_paths]
        merge_io(infiles, outfile, bufsize=bufsize)


def main() -> None:
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        required=True,
        dest="command",
        help=(
            "Commands:\n"
            "split: split a file into multiple files\n"
            "merge: merge multiple files into a file"
            "Example:\n"
            "tally-token split example.bin example.bin.1 example.bin.2 example.bin.3\n"
            "tally-token merge example-merged.bin example.bin.1 example.bin.2 example.bin.3"
        ),
    )
    split_parser = subparsers.add_parser("split")
    split_parser.add_argument("src", help="The source file to be split.")
    split_parser.add_argument("dst", nargs="+", help="The destination files.")
    split_parser.add_argument(
        "--bufsize", type=int, default=1024 * 2, help="The buffer size."
    )

    merge_parser = subparsers.add_parser("merge")
    merge_parser.add_argument("dst", help="The destination file.")
    merge_parser.add_argument("src", nargs="+", help="The source files.")
    merge_parser.add_argument(
        "--bufsize", type=int, default=1024**2, help="The buffer size."
    )

    args = parser.parse_args()
    if args.command == "split":
        split_main(source_path=args.src, dest_paths=args.dst)
    elif args.command == "merge":
        merge_main(dest_path=args.dst, source_paths=args.src)


if __name__ == "__main__":
    main()
